use aoc_runner_derive::{aoc, aoc_generator};
use hashbrown::HashMap;
use nom::{
    character::complete::{space1, u64 as parse_usize},
    error::Error,
    multi::separated_list1,
};

type Input = Vec<usize>;

#[aoc_generator(day11)]
fn parse(input: &str) -> Input {
    separated_list1(space1::<&str, Error<&str>>, parse_usize)(input)
        .unwrap()
        .1
        .into_iter()
        .map(|v| v as usize)
        .collect()
}

#[aoc(day11, part1)]
fn part1(input: &Input) -> usize {
    blink(25, input.clone())
}

#[aoc(day11, part2)]
fn part2(input: &Input) -> usize {
    blink(75, input.clone())
}

fn blink(blinks: usize, input: Vec<usize>) -> usize {
    let mut stones = HashMap::new();
    for stone in input {
        *stones.entry(stone).or_insert(0) += 1;
    }
    for _ in 0..blinks {
        stones = single_blink(stones);
    }
    stones.values().sum::<usize>()
}

fn single_blink(stones: HashMap<usize, usize>) -> HashMap<usize, usize> {
    let mut new_stones = HashMap::with_capacity(2 << 12);

    for (stone, count) in stones {
        if stone == 0 {
            *new_stones.entry(1).or_insert(0) += count;
        } else {
            let digits = stone.ilog10() + 1;
            if digits % 2 == 0 {
                let left = stone / 10usize.pow(digits / 2);
                let right = stone % 10usize.pow(digits / 2);
                *new_stones.entry(left).or_insert(0) += count;
                *new_stones.entry(right).or_insert(0) += count;
            } else {
                *new_stones.entry(stone * 2024).or_insert(0) += count;
            }
        }
    }

    new_stones
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn part1_example() {
        assert_eq!(
            part1(&parse(include_str!("../input/2024/day11.txt").trim_end())),
            203609
        );
    }

    #[test]
    fn part2_example() {
        assert_eq!(
            part2(&parse(include_str!("../input/2024/day11.txt").trim_end())),
            240954878211138
        );
    }
}
