use aoc_runner_derive::{aoc, aoc_generator};
use nom::{
    bytes::complete::tag,
    character::complete::{newline, one_of},
    combinator::map,
    multi::{many1, separated_list1},
    sequence::separated_pair,
    IResult,
};

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum Color {
    White,
    Blue,
    Black,
    Red,
    Green,
}

type Input = (Vec<Vec<Color>>, Vec<Vec<Color>>);

#[aoc_generator(day19)]
fn parse(input: &str) -> Input {
    separated_pair(
        separated_list1(tag(", "), many1(parse_color)),
        tag("\n\n"),
        separated_list1(newline, many1(parse_color)),
    )(input)
    .unwrap()
    .1
}

fn parse_color(input: &str) -> IResult<&str, Color> {
    map(one_of("wubrg"), |c| match c {
        'w' => Color::White,
        'u' => Color::Blue,
        'b' => Color::Black,
        'r' => Color::Red,
        'g' => Color::Green,
        _ => unreachable!(),
    })(input)
}

#[aoc(day19, part1)]
fn part1(input: &Input) -> usize {
    let towels = &input.0;
    let patterns = &input.1;

    patterns
        .iter()
        .map(|pattern| {
            let mut m = vec![0u64; pattern.len() + 1];
            m[0] = 1;

            for i in 1..=pattern.len() {
                for towel in towels {
                    if towel.len() <= i && &pattern[i - towel.len()..i] == towel {
                        m[i] += m[i - towel.len()];
                    }
                }
            }

            m[pattern.len()]
        })
        .filter(|n| *n > 0)
        .count()
}

#[aoc(day19, part2)]
fn part2(input: &Input) -> u64 {
    let towels = &input.0;
    let patterns = &input.1;

    patterns
        .iter()
        .map(|pattern| {
            let mut m = vec![0u64; pattern.len() + 1];
            m[0] = 1;

            for i in 1..=pattern.len() {
                for towel in towels {
                    if towel.len() <= i && &pattern[i - towel.len()..i] == towel {
                        m[i] += m[i - towel.len()];
                    }
                }
            }

            m[pattern.len()]
        })
        .sum()
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn part1_example() {
        assert_eq!(
            part1(&parse(
                "r, wr, b, g, bwu, rb, gb, br

brwrr
bggr
gbbr
rrbgbr
ubwu
bwurrg
brgr
bbrgwb"
            )),
            6
        );
    }

    // #[test]
    // fn part2_example() {
    //     assert_eq!(part2(&parse("<EXAMPLE>")), 0);
    // }
}
