use aoc_runner_derive::{aoc, aoc_generator};
#[aoc_generator(day4)]
fn parse(input: &str) -> Vec<Vec<char>> {
    input
        .split("\n")
        .map(|line| line.chars().collect())
        .collect()
}

#[aoc(day4, part1)]
fn part1(d: &[Vec<char>]) -> usize {
    (0..d.len()).fold(0, |z, i| {
        z + (0..d[0].len())
            .map(move |j| {
                (
                    j as i64,
                    i as i64,
                    j > 2,
                    j < d.len() - 3,
                    i > 2,
                    i < d[0].len() - 3,
                )
            })
            .fold(0, |z, p| {
                z + (-1..=1).fold(0, |z, f| {
                    z + (-1..=1).fold(0, move |z, g| {
                        z + ((f != 0 || g != 0)
                            && (f == 0 || (f == -1 && p.2) || (f == 1 && p.3))
                            && (g == 0 || (g == -1 && p.4) || (g == 1 && p.5))
                            && (0..=3).zip("XMAS".chars()).all(|v| {
                                d[(p.1 + v.0 * g) as usize][(p.0 + v.0 * f) as usize] == v.1
                            })) as usize
                    })
                })
            })
    })
}

fn _part1(input: &[Vec<char>]) -> u64 {
    let h = input.len();
    let w = input[0].len();
    let mut sum = 0;
    for y in 0..h {
        for x in 0..w {
            if input[y][x] == 'S' {
                if x < w - 3
                    && input[y][x + 1] == 'A'
                    && input[y][x + 2] == 'M'
                    && input[y][x + 3] == 'X'
                {
                    // println!("rev X {} {}", x, y);
                    sum += 1
                }
                if y < h - 3
                    && input[y + 1][x] == 'A'
                    && input[y + 2][x] == 'M'
                    && input[y + 3][x] == 'X'
                {
                    // println!("rev Y {} {}", x, y);
                    sum += 1
                }
                if x < w - 3
                    && y < h - 3
                    && input[y + 1][x + 1] == 'A'
                    && input[y + 2][x + 2] == 'M'
                    && input[y + 3][x + 3] == 'X'
                {
                    // println!("rdi Y {} {}", x, y);
                    sum += 1
                }
                if x >= 3
                    && y < h - 3
                    && input[y + 1][x - 1] == 'A'
                    && input[y + 2][x - 2] == 'M'
                    && input[y + 3][x - 3] == 'X'
                {
                    // println!("rdi Y {} {}", x, y);
                    sum += 1
                }
            } else if input[y][x] == 'X' {
                if x < w - 3
                    && input[y][x + 1] == 'M'
                    && input[y][x + 2] == 'A'
                    && input[y][x + 3] == 'S'
                {
                    // println!("    X {} {}", x, y);
                    sum += 1
                }
                if y < h - 3
                    && input[y + 1][x] == 'M'
                    && input[y + 2][x] == 'A'
                    && input[y + 3][x] == 'S'
                {
                    // println!("    Y {} {}", x, y);
                    sum += 1
                }
                if x < w - 3
                    && y < h - 3
                    && input[y + 1][x + 1] == 'M'
                    && input[y + 2][x + 2] == 'A'
                    && input[y + 3][x + 3] == 'S'
                {
                    // println!("dia Y {} {}", x, y);
                    sum += 1
                }
                if x >= 3
                    && y < h - 3
                    && input[y + 1][x - 1] == 'M'
                    && input[y + 2][x - 2] == 'A'
                    && input[y + 3][x - 3] == 'S'
                {
                    // println!("dia Y {} {}", x, y);
                    sum += 1
                }
            }
        }
    }
    sum
}

#[aoc(day4, part2)]
fn part2(input: &[Vec<char>]) -> u64 {
    let h = input.len();
    let w = input[0].len();
    let mut sum = 0;
    for y in 1..h - 1 {
        for x in 1..w - 1 {
            if input[y][x] == 'A' {
                let diag1 = (input[y - 1][x - 1] == 'M' && input[y + 1][x + 1] == 'S')
                    || (input[y - 1][x - 1] == 'S' && input[y + 1][x + 1] == 'M');
                let diag2 = (input[y - 1][x + 1] == 'M' && input[y + 1][x - 1] == 'S')
                    || (input[y - 1][x + 1] == 'S' && input[y + 1][x - 1] == 'M');
                if diag1 && diag2 {
                    sum += 1;
                }
            }
        }
    }
    sum
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn part1_example() {
        assert_eq!(part1(&parse("MMMSXXMASM\nMSAMXMSMSA\nAMXSXMAAMM\nMSAMASMSMX\nXMASAMXAMM\nXXAMMXXAMA\nSMSMSASXSS\nSAXAMASAAA\nMAMMMXMMMM\nMXMXAXMASX")), 18);
        assert_eq!(part1(&parse(include_str!("../input/2024/day4.txt"))), 2406);
    }

    #[test]
    fn part2_example() {
        assert_eq!(part2(&parse(".M.S......\n..A..MSMS.\n.M.S.MAA..\n..A.ASMSM.\n.M.S.M....\n..........\nS.S.S.S.S.\n.A.A.A.A..\nM.M.M.M.M.\n..........")), 9);
    }
}
