use hashbrown::HashMap;

use aoc_runner_derive::{aoc, aoc_generator};

type Input = ((isize, isize), Vec<Vec<(isize, isize)>>);

#[aoc_generator(day8)]
fn parse(input: &str) -> Input {
    let mut map = HashMap::new();
    let h = input.lines().count();
    let w = input.lines().find(|_| true).unwrap().len();
    input.lines().enumerate().for_each(|(y, line)| {
        line.chars()
            .enumerate()
            .filter(|(_, c)| c != &'.')
            .for_each(|(x, c)| {
                map.entry(c)
                    .or_insert(Vec::new())
                    .push((x as isize, y as isize));
            });
    });
    let vec = map.into_values().collect();
    ((w as isize, h as isize), vec)
}

#[aoc(day8, part1)]
fn part1(input: &Input) -> usize {
    let mut set = [[false; 64]; 64];
    let mut sum = 0;
    input.1.iter().for_each(|vec| {
        for l in 0..vec.len() {
            for r in l + 1..vec.len() {
                let a1 = vec[l];
                let a2 = vec[r];
                let c1 = (2 * a1.0 - a2.0, 2 * a1.1 - a2.1);
                let c2 = (2 * a2.0 - a1.0, 2 * a2.1 - a1.1);
                if c1.0 >= 0
                    && c1.0 < input.0 .0
                    && c1.1 >= 0
                    && c1.1 < input.0 .1
                    && !set[c1.1 as usize][c1.0 as usize]
                {
                    set[c1.1 as usize][c1.0 as usize] = true;
                    sum += 1;
                }
                if c2.0 >= 0
                    && c2.0 < input.0 .0
                    && c2.1 >= 0
                    && c2.1 < input.0 .1
                    && !set[c2.1 as usize][c2.0 as usize]
                {
                    set[c2.1 as usize][c2.0 as usize] = true;
                    sum += 1;
                }
            }
        }
    });
    sum
}

#[aoc(day8, part2)]
fn part2(input: &Input) -> usize {
    let mut set = [[false; 64]; 64];
    let mut sum = 0;
    input.1.iter().for_each(|vec| {
        for l in 0..vec.len() {
            for r in l + 1..vec.len() {
                let a1 = vec[l];
                let a2 = vec[r];
                let s = a1;
                let v = (a2.0 - a1.0, a2.1 - a1.1);
                let mut d = 0;
                loop {
                    let c = (s.0 + d * v.0, s.1 + d * v.1);
                    d += 1;
                    if c.0 >= 0 && c.0 < input.0 .0 && c.1 >= 0 && c.1 < input.0 .1 {
                        if !set[c.1 as usize][c.0 as usize] {
                            set[c.1 as usize][c.0 as usize] = true;
                            sum += 1;
                        };
                    } else {
                        break;
                    }
                }
                let mut d = -1;
                loop {
                    let c = (s.0 + d * v.0, s.1 + d * v.1);
                    d -= 1;
                    if c.0 >= 0 && c.0 < input.0 .0 && c.1 >= 0 && c.1 < input.0 .1 {
                        if !set[c.1 as usize][c.0 as usize] {
                            set[c.1 as usize][c.0 as usize] = true;
                            sum += 1;
                        };
                    } else {
                        break;
                    }
                }
            }
        }
    });
    sum
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn part1_example() {
        assert_eq!(
            part1(&parse(
                "............\n........0...\n.....0......\n.......0....\n....0.......\n......A.....\n............\n............\n........A...\n.........A..\n............\n............"
            )),
            14
        );
        assert_eq!(
            part1(&parse(include_str!("../input/2024/day8.txt").trim_end())),
            273
        );
    }

    #[test]
    fn part2_example() {
        assert_eq!(
            part2(&parse(
                "............\n........0...\n.....0......\n.......0....\n....0.......\n......A.....\n............\n............\n........A...\n.........A..\n............\n............"
            )),
            34
        );
        assert_eq!(
            part2(&parse(include_str!("../input/2024/day8.txt").trim_end())),
            1017
        );
    }
}
